"""Generate a dependency tree."""


from __future__ import annotations

import logging

from typing import TYPE_CHECKING, Union

from pip4a.tree import Tree
from pip4a.utils import collect_manifests


if TYPE_CHECKING:
    from pip4a.config import Config


logger = logging.getLogger(__name__)

ScalarVal = Union[bool, str, float, int, None]
JSONVal = Union[ScalarVal, list["JSONVal"], dict[str, "JSONVal"]]


class TreeMaker:
    """Generate a dependency tree."""

    def __init__(self: TreeMaker, config: Config) -> None:
        """Initialize the object."""
        self._config = config

    def run(self: TreeMaker) -> None:  # noqa: C901, PLR0912, PLR0915
        """Run the command."""
        # pylint: disable=too-many-locals
        collections = collect_manifests(
            target=self._config.site_pkg_collections_path,
            venv_cache_dir=self._config.venv_cache_dir,
        )
        tree_dict: JSONVal = {c: {} for c in collections}
        if not isinstance(tree_dict, dict):
            msg = "Tree dict is not a dict."
            raise TypeError(msg)

        links: dict[str, str] = {}
        for collection_name, collection in collections.items():
            err = f"Collection {collection_name} has malformed metadata."
            if not isinstance(collection["collection_info"], dict):
                logger.error(err)
                continue
            if not isinstance(collection["collection_info"]["dependencies"], dict):
                logger.error(err)
                continue

            for dep in collection["collection_info"]["dependencies"]:
                if not isinstance(dep, str):
                    err = f"Collection {collection_name} has malformed dependency."
                    logger.error(err)
                    continue
                target = tree_dict[collection_name]
                if not isinstance(target, dict):
                    msg = "Tree dict is not a dict."
                    raise TypeError(msg)

                target[dep] = tree_dict[dep]

            docs = collection["collection_info"].get("documentation")
            homepage = collection["collection_info"].get("homepage")
            repository = collection["collection_info"].get("repository")
            issues = collection["collection_info"].get("issues")
            link = docs or homepage or repository or issues or "http://ansible.com"
            if not isinstance(link, str):
                msg = "Link is not a string."
                raise TypeError(msg)
            links[collection_name] = link

        if self._config.args.verbose:
            tree = Tree(obj=tree_dict, term_features=self._config.term_features)
            tree.links = links
            rendered = tree.render()
            print(rendered)  # noqa: T201
            return

        pruned_tree_dict: JSONVal = {}
        if not isinstance(pruned_tree_dict, dict):
            msg = "Tree dict is not a dict."
            raise TypeError(msg)
        for collection_name in list(tree_dict.keys()):
            found = False
            for value in tree_dict.values():
                if not isinstance(value, dict):
                    msg = "Tree dict is not a dict."
                    raise TypeError(msg)
                if collection_name in value:
                    found = True
            if not found:
                pruned_tree_dict[collection_name] = tree_dict[collection_name]

        tree = Tree(obj=pruned_tree_dict, term_features=self._config.term_features)
        tree.links = links
        rendered = tree.render()
        print(rendered)  # noqa: T201



"""Produce pretty logs."""

from __future__ import annotations

import logging

from copy import copy


MAPPING = {
    "DEBUG": "30",  # grey
    "INFO": "92",  # bright cyan
    "WARNING": "93",  # bright yellow
    "ERROR": "31;1",  # bold red
    "CRITICAL": "91;1",  # bold bright red
}

PREFIX = "\033["
SUFFIX = "\033[0m"


class ColoredFormatter(logging.Formatter):
    """A colorful formatter."""

    def __init__(self: ColoredFormatter, pattern: str) -> None:
        """Initialize the formatter.

        Args:
            pattern: The log format
        """
        logging.Formatter.__init__(self, pattern)

    def format(  # noqa: A003
        self: ColoredFormatter,
        record: logging.LogRecord,
    ) -> str:
        """Format the log record.

        Args:
            record: The log record

        Returns:
            The formatted log record
        """
        colored_record = copy(record)
        levelname = colored_record.levelname
        seq = MAPPING.get(levelname, 37)  # default white
        colored_levelname = (
            f"{PREFIX}{seq}m{levelname.lower().capitalize() + ':':<9}{SUFFIX}"
        )
        colored_record.levelname = colored_levelname
        colored_msg = f"{PREFIX}{seq}m{record.msg}{SUFFIX}"
        colored_record.msg = colored_msg
        return logging.Formatter.format(self, colored_record)


class ExitOnExceptionHandler(logging.StreamHandler):  # type: ignore[type-arg]
    """Exit on exception handler."""

    def emit(self: ExitOnExceptionHandler, record: logging.LogRecord) -> None:
        """Emit the log record.

        Args:
            record: The log record

        Raises:
            SystemExit: If the log record is an error or critical
        """
        super().emit(record)
        if record.levelno == logging.CRITICAL:
            raise SystemExit(1)



"""The installer."""

from __future__ import annotations

import logging
import shutil

from pathlib import Path
from typing import TYPE_CHECKING

from pip4a.utils import collections_from_requirements, note, parse_collection_request


if TYPE_CHECKING:
    from pip4a.config import Config


logger = logging.getLogger(__name__)


class UnInstaller:
    """The uninstaller class."""

    def __init__(self: UnInstaller, config: Config) -> None:
        """Initialize the installer.

        Args:
            config: The application configuration.
        """
        self._config = config

    def run(self: UnInstaller) -> None:
        """Run the uninstaller."""
        if self._config.args.requirement:
            requirements_path = Path(self._config.args.requirement)
            if not requirements_path.exists():
                err = f"Failed to find requirements file: {requirements_path}"
                logger.critical(err)
            collections = collections_from_requirements(requirements_path)
            for collection in collections:
                self._config.collection = parse_collection_request(
                    collection["name"],
                )
                self._remove_collection()
        else:
            self._remove_collection()

    def _remove_collection(self: UnInstaller) -> None:
        """Remove the collection."""
        msg = f"Checking {self._config.collection.name} at {self._config.site_pkg_collection_path}"
        logger.debug(msg)

        if self._config.site_pkg_collection_path.exists():
            msg = f"Exists: {self._config.site_pkg_collection_path}"
            logger.debug(msg)

            if self._config.site_pkg_collection_path.is_symlink():
                self._config.site_pkg_collection_path.unlink()
            else:
                shutil.rmtree(self._config.site_pkg_collection_path)
            msg = f"Removed {self._config.collection.name}"
            note(msg)
        else:
            err = (
                f"Failed to find {self._config.collection.name}:"
                f" {self._config.site_pkg_collection_path}"
            )
            logger.warning(err)

        for entry in self._config.site_pkg_collections_path.iterdir():
            if all(
                (
                    entry.is_dir(),
                    entry.name.startswith(self._config.collection.name),
                    entry.suffix == ".info",
                ),
            ):
                shutil.rmtree(entry)
                msg = f"Removed {self._config.collection.name}*.info: {entry}"
                logger.debug(msg)

        collection_namespace_root = self._config.site_pkg_collection_path.parent
        try:
            collection_namespace_root.rmdir()
            msg = f"Removed collection namespace root: {collection_namespace_root}"
            logger.debug(msg)
        except FileNotFoundError:
            pass
        except OSError as exc:
            msg = f"Failed to remove collection namespace root: {exc}"
            logger.debug(msg)

        try:
            self._config.site_pkg_collections_path.rmdir()
            msg = f"Removed collection root: {self._config.site_pkg_collections_path}"
            logger.debug(msg)
        except FileNotFoundError:
            pass
        except OSError as exc:
            msg = f"Failed to remove collection root: {exc}"
            logger.debug(msg)



"""Unit test for the utilities module."""

from __future__ import annotations

from pathlib import Path

import pytest

from pip4a.utils import CollectionSpec, parse_collection_request


scenarios = (
    ("ansible.utils", CollectionSpec(cname="utils", cnamespace="ansible", local=False)),
    (
        "ansible.utils:1.0.0",
        CollectionSpec(
            cname="utils",
            cnamespace="ansible",
            specifier=":1.0.0",
            local=False,
        ),
    ),
    (
        "ansible.utils>=1.0.0",
        CollectionSpec(
            cname="utils",
            cnamespace="ansible",
            specifier=">=1.0.0",
            local=False,
        ),
    ),
    (
        "/",
        CollectionSpec(
            specifier=None,
            local=True,
            path=Path("/"),
        ),
    ),
    (
        "/[test]",
        CollectionSpec(
            specifier=None,
            local=True,
            path=Path("/"),
            opt_deps="test",
        ),
    ),
    (
        "/foo/bar",
        None,
    ),
    (
        "abcdefg",
        None,
    ),
)


@pytest.mark.parametrize("scenario", scenarios)
def test_parse_collection_request(scenario: tuple[str, CollectionSpec | None]) -> None:
    """Test that the parse_collection_request function works as expected."""
    string, spec = scenario
    if spec is None:
        with pytest.raises(SystemExit):
            parse_collection_request(string)
    else:
        assert parse_collection_request(string) == spec



"""Lister module for pip4a."""

from __future__ import annotations

import logging

from typing import TYPE_CHECKING

from pip4a.utils import collect_manifests, term_link


if TYPE_CHECKING:
    from pip4a.config import Config


logger = logging.getLogger(__name__)


class Lister:
    """The Lister class."""

    def __init__(self: Lister, config: Config, output_format: str) -> None:
        """Initialize the Lister.

        Args:
            config: The application configuration.
            output_format: The output format to use.
        """
        self._config: Config = config
        self._output_format: str = output_format

    def run(self: Lister) -> None:
        """Run the Lister."""
        # pylint: disable=too-many-locals
        collections = collect_manifests(
            target=self._config.site_pkg_collections_path,
            venv_cache_dir=self._config.venv_cache_dir,
        )

        if self._output_format == "list":
            column1_width = 30
            column2_width = 10
            column3_width = 25

            print(  # noqa: T201
                f"{'Collection': <{column1_width}}"
                f" {'Version': <{column2_width}}"
                f" {'Editable project location': <{column3_width}}",
            )
            print(  # noqa: T201
                f"{'-' * (column1_width)}"
                f" {'-' * (column2_width)}"
                f" {'-' * (column3_width)}",
            )

            for fqcn, collection in collections.items():
                err = f"Collection {fqcn} has malformed metadata."
                ci = collection["collection_info"]
                if not isinstance(ci, dict):
                    logger.error(err)
                    continue
                collection_name = ci["name"]
                collection_namespace = ci["namespace"]
                collection_version = ci["version"]
                if not isinstance(collection_name, str):
                    logger.error(err)
                    continue
                if not isinstance(collection_namespace, str):
                    logger.error(err)
                    continue
                if not isinstance(collection_version, str):
                    logger.error(err)
                    continue

                collection_path = (
                    self._config.site_pkg_collections_path
                    / collection_namespace
                    / collection_name
                )
                if collection_path.is_symlink():
                    editable_location = str(collection_path.resolve())
                else:
                    editable_location = ""

                docs = ci.get("documentation")
                homepage = ci.get("homepage")
                repository = ci.get("repository")
                issues = ci.get("issues")
                link = docs or homepage or repository or issues or "http://ansible.com"
                if not isinstance(link, str):
                    msg = "Link is not a string."
                    raise TypeError(msg)
                fqcn_linked = term_link(
                    uri=link,
                    label=fqcn,
                    term_features=self._config.term_features,
                )

                print(  # noqa: T201
                    fqcn_linked + " " * (column1_width - len(fqcn)),
                    f"{ci['version']: <{column2_width}}",
                    f"{editable_location: <{column3_width}}",
                )



"""CLI entrypoint."""

from __future__ import annotations

import logging
import os
import site
import sys

from pathlib import Path
from typing import TYPE_CHECKING

from .arg_parser import parse
from .config import Config
from .logger import ColoredFormatter, ExitOnExceptionHandler
from .subcommands.checker import Checker
from .subcommands.inspector import Inspector
from .subcommands.installer import Installer
from .subcommands.lister import Lister
from .subcommands.treemaker import TreeMaker
from .subcommands.uninstaller import UnInstaller
from .utils import TermFeatures


if TYPE_CHECKING:
    from argparse import Namespace


class LogType:
    """A log type."""

    def __init__(self: LogType) -> None:
        """Initialize the log type."""
        self.errors = 0


class LogCount(logging.Handler):
    """A log counter."""

    def __init__(self: LogCount) -> None:
        """Initialize the log counter."""
        super().__init__()
        self.count = LogType()
        self.name = "counter"

    def emit(self: LogCount, record: logging.LogRecord) -> None:
        """Emit the log record.

        Args:
            record: The log record
        """
        if record.levelname == "ERROR":
            self.count.errors += 1


class Cli:
    """The Cli class."""

    def __init__(self: Cli) -> None:
        """Initialize the CLI and parse CLI args."""
        self.args: Namespace
        self.config: Config
        self.logger: logging.Logger

    def parse_args(self: Cli) -> None:
        """Parse the command line arguments."""
        self.args = parse()
        if hasattr(self.args, "requirement") and self.args.requirement:
            self.args.requirement = Path(self.args.requirement).expanduser().resolve()

    def init_logger(self: Cli) -> None:
        """Initialize the logger."""
        self.logger = logging.getLogger("pip4a")
        count = LogCount()
        self.logger.addHandler(count)

        ch = ExitOnExceptionHandler()
        cf = ColoredFormatter(
            "%(levelname)s %(message)s",
        )
        ch.setFormatter(cf)
        self.logger.addHandler(ch)

        max_verbosity = 3
        if self.args.verbose > max_verbosity:
            self.args.verbose = 3

        log_level = logging.ERROR - (self.args.verbose * 10)
        self.logger.setLevel(log_level)

    def args_sanity(self: Cli) -> None:
        """Perform some sanity checking on the args."""
        if (
            hasattr(self.args, "requirement")
            and self.args.requirement
            and not self.args.requirement.exists()
        ):
            err = f"Requirements file not found: {self.args.requirement}"
            self.logger.critical(err)

    def ensure_isolated(self: Cli) -> None:
        """Ensure the environment is isolated."""
        env_vars = os.environ
        errors = []
        if "ANSIBLE_COLLECTIONS_PATHS" in env_vars:
            err = "ANSIBLE_COLLECTIONS_PATHS is set"
            errors.append(err)
        if "ANSIBLE_COLLECTION_PATH" in env_vars:
            err = "ANSIBLE_COLLECTION_PATH is set"
            errors.append(err)

        home_coll = Path.home() / ".ansible/collections/ansible_collections"
        if home_coll.exists() and tuple(home_coll.iterdir()):
            err = f"Collections found in {home_coll}"
            errors.append(err)

        usr_coll = Path("/usr/share/ansible/collections")
        if usr_coll.exists() and tuple(usr_coll.iterdir()):
            err = f"Collections found in {usr_coll}"
            errors.append(err)

        if "VIRTUAL_ENV" not in env_vars and not self.args.venv:
            err = (
                "Unable to use user site packages directory:"
                f" {site.getusersitepackages()}, please activate or specify a virtual environment"
            )
            errors.append(err)

        if errors:
            err = (
                "The development environment is not isolated,"
                " please resolve the following errors:"
            )
            self.logger.error(err)
            for error in errors:
                err = f"- {error}"
                self.logger.error(err)
            err = "Exiting."
            self.logger.critical(err)

    def run(self: Cli) -> None:
        """Run the application."""
        logging.getLogger("pip4a")

        term_features = TermFeatures(
            color=False if os.environ.get("NO_COLOR") else not self.args.no_ansi,
            links=not self.args.no_ansi,
        )

        self.config = Config(args=self.args, term_features=term_features)

        if self.config.args.subcommand == "check":
            self.config.init()
            checker = Checker(config=self.config)
            checker.run()
            self._exit()

        if self.config.args.subcommand == "inspect":
            self.config.init()
            inspector = Inspector(config=self.config)
            inspector.run()
            self._exit()

        if self.config.args.subcommand == "install":
            self.config.init(create_venv=True)
            installer = Installer(self.config)
            installer.run()
            self._exit()

        if self.config.args.subcommand == "list":
            self.config.init()
            lister = Lister(config=self.config, output_format="list")
            lister.run()
            self._exit()

        if self.config.args.subcommand == "tree":
            self.config.init()
            treemaker = TreeMaker(self.config)
            treemaker.run()
            self._exit()

        if self.config.args.subcommand == "uninstall":
            self.config.init()
            uninstaller = UnInstaller(self.config)
            uninstaller.run()
            self._exit()

    def _exit(self: Cli) -> None:
        """Exit the application setting the return code."""
        logger = logging.getLogger("pip4a")
        status = 0
        for handler in logger.handlers:
            if handler.name == "counter":
                status = int(bool(handler.count.errors))  # type: ignore[attr-defined]
        sys.exit(status)


def main() -> None:
    """Entry point for ansible-creator CLI."""
    cli = Cli()
    cli.parse_args()
    cli.init_logger()
    cli.args_sanity()
    cli.ensure_isolated()
    cli.run()


if __name__ == "__main__":
    main()



"""The collection inspect command."""

from __future__ import annotations

import json
import logging
import os

from typing import TYPE_CHECKING

from pip4a.utils import collect_manifests


if TYPE_CHECKING:
    from pip4a.config import Config

try:
    from pip._vendor.rich import print_json

    HAS_RICH = True
except ImportError:
    HAS_RICH = False


logger = logging.getLogger(__name__)


class Inspector:
    """The Inspector class."""

    def __init__(self: Inspector, config: Config) -> None:
        """Initialize the Inspector."""
        self._config: Config = config

    def run(self: Inspector) -> None:
        """Run the Inspector."""
        # pylint: disable=too-many-locals
        collections = collect_manifests(
            target=self._config.site_pkg_collections_path,
            venv_cache_dir=self._config.venv_cache_dir,
        )

        output = json.dumps(collections, indent=4, sort_keys=True)
        if HAS_RICH and not os.environ.get("NOCOLOR"):
            print_json(output)
        else:
            print(output)  # noqa: T201



"""Parse the command line arguments."""
import argparse


try:
    from ._version import version as __version__
except ImportError:  # pragma: no cover
    try:
        import pkg_resources

        __version__ = pkg_resources.get_distribution("pip4a").version
    except Exception:  # pylint: disable=broad-except # noqa: BLE001
        # this is the fallback SemVer version picked by setuptools_scm when tag
        # information is not available.
        __version__ = "0.1.dev1"


def parse() -> argparse.Namespace:
    """Parse the command line arguments.

    Returns:
        The arguments
    """
    parser = argparse.ArgumentParser(
        description="A pip-like ansible collection installer.",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.add_argument(
        "--version",
        action="version",
        help="Version specifier.",
        version=__version__,
    )

    subparsers = parser.add_subparsers(
        title="subcommands",
        description="valid subcommands",
        dest="subcommand",
        required=True,
    )

    level1 = argparse.ArgumentParser(add_help=False)
    level1.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="Give more output. Option is additive, and can be used up to 3 times.",
    )
    level1.add_argument(
        "--venv",
        help="Target virtual environment.",
    )

    level1.add_argument(
        "--no-ansi",
        action="store_true",
        default=False,
        help="Disable the use of ANSI codes for terminal hyperlink generation and color.",
    )

    _check = subparsers.add_parser(
        "check",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        parents=[level1],
        help="Check installed collections",
    )

    _inspect = subparsers.add_parser(
        "inspect",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        parents=[level1],
        help="Inspect installed collections",
    )

    _list = subparsers.add_parser(
        "list",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        parents=[level1],
        help="List installed collections",
    )

    _tree = subparsers.add_parser(
        "tree",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        parents=[level1],
        help="Generate a dependency tree",
    )

    level2 = argparse.ArgumentParser(add_help=False, parents=[level1])
    spec_or_req = level2.add_mutually_exclusive_group(required=True)
    spec_or_req.add_argument(
        "collection_specifier",
        help="Collection name or path to collection with extras.",
        nargs="?",
    )
    spec_or_req.add_argument(
        "-r",
        "--requirement",
        help="Install from the given requirements file.",
        required=False,
    )

    install_usage = """Usage:
        pip4a install .
        pip4a install -e .
        pip4a install -e .[test]
        python -m pip4a install ansible.utils"""

    install = subparsers.add_parser(
        "install",
        epilog=install_usage,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        parents=[level2],
        help="Install a collection.",
    )
    install.add_argument(
        "-e",
        "--editable",
        action="store_true",
        help="Install editable.",
    )

    _uninstall = subparsers.add_parser(
        "uninstall",
        epilog=install_usage,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        parents=[level2],
        help="Uninstall a collection.",
    )

    return parser.parse_args()



"""The installer."""

from __future__ import annotations

import logging
import re
import shutil
import subprocess

from pathlib import Path
from typing import TYPE_CHECKING

from pip4a.utils import (
    builder_introspect,
    collections_from_requirements,
    note,
    oxford_join,
    subprocess_run,
)


if TYPE_CHECKING:
    from pip4a.config import Config


logger = logging.getLogger(__name__)


class Installer:
    """The installer class."""

    def __init__(self: Installer, config: Config) -> None:
        """Initialize the installer.

        Args:
            config: The application configuration.
        """
        self._config = config

    def run(self: Installer) -> None:
        """Run the installer."""
        if self._config.args.editable and not self._config.collection.local:
            err = "Editable installs are only supported for local collections."
            logger.critical(err)

        if (
            self._config.args.collection_specifier
            and "," in self._config.args.collection_specifier
        ):
            err = "Multiple optional dependencies are not supported at this time."
            logger.critical(err)

        self._install_core()

        if self._config.args.requirement:
            self._install_galaxy_requirements()
        elif self._config.collection.local:
            self._install_local_collection()
            if self._config.args.editable:
                self._swap_editable_collection()
        elif not self._config.collection.local:
            self._install_galaxy_collection()

        builder_introspect(config=self._config)
        self._pip_install()
        self._check_bindep()

        if self._config.args.venv and (
            self._config.interpreter != self._config.venv_interpreter
        ):
            msg = "A virtual environment was specified but has not been activated."
            note(msg)
            msg = (
                "Please activate the virtual environment:"
                f"\nsource {self._config.args.venv}/bin/activate"
            )
            note(msg)

    def _install_core(self: Installer) -> None:
        """Install ansible-core if not installed already."""
        msg = "Installing ansible-core."
        logger.info(msg)

        core = self._config.venv_bindir / "ansible"
        if core.exists():
            return
        msg = "Installing ansible-core."
        logger.debug(msg)
        command = f"{self._config.venv_interpreter} -m pip install ansible-core"
        try:
            subprocess_run(command=command, verbose=self._config.args.verbose)
        except subprocess.CalledProcessError as exc:
            err = f"Failed to install ansible-core: {exc}"
            logger.critical(err)

    def _install_galaxy_collection(self: Installer) -> None:
        """Install the collection from galaxy."""
        msg = f"Installing collection from galaxy: {self._config.args.collection_specifier}"
        logger.info(msg)

        if self._config.site_pkg_collection_path.exists():
            msg = f"Removing installed {self._config.site_pkg_collection_path}"
            logger.debug(msg)
            if self._config.site_pkg_collection_path.is_symlink():
                self._config.site_pkg_collection_path.unlink()
            else:
                shutil.rmtree(self._config.site_pkg_collection_path)

        command = (
            f"{self._config.venv_bindir / 'ansible-galaxy'} collection"
            f" install '{self._config.args.collection_specifier}'"
            f" -p {self._config.site_pkg_path}"
            " --force"
        )
        env = {
            "ANSIBLE_GALAXY_COLLECTIONS_PATH_WARNING": str(self._config.args.verbose),
        }
        msg = "Running ansible-galaxy to install non-local collection and it's dependencies."
        logger.debug(msg)
        try:
            proc = subprocess_run(
                command=command,
                env=env,
                verbose=self._config.args.verbose,
            )
        except subprocess.CalledProcessError as exc:
            err = f"Failed to install collection: {exc} {exc.stderr}"
            logger.critical(err)
            return
        installed = re.findall(r"(\w+\.\w+):.*installed", proc.stdout)
        msg = f"Installed collections include: {oxford_join(installed)}"
        note(msg)

    def _install_galaxy_requirements(self: Installer) -> None:
        """Install the collections using requirements.yml."""
        msg = f"Installing collections from requirements file: {self._config.args.requirement}"
        logger.info(msg)

        collections = collections_from_requirements(file=self._config.args.requirement)
        for collection in collections:
            cnamespace = collection["name"].split(".")[0]
            cname = collection["name"].split(".")[1]
            cpath = self._config.site_pkg_collections_path / cnamespace / cname
            if cpath.exists():
                msg = f"Removing installed {cpath}"
                logger.debug(msg)
                if cpath.is_symlink():
                    cpath.unlink()
                else:
                    shutil.rmtree(cpath)

        command = (
            f"{self._config.venv_bindir / 'ansible-galaxy'} collection"
            f" install -r {self._config.args.requirement}"
            f" -p {self._config.site_pkg_path}"
            " --force"
        )
        try:
            proc = subprocess_run(command=command, verbose=self._config.args.verbose)
        except subprocess.CalledProcessError as exc:
            err = f"Failed to install collections: {exc} {exc.stderr}"
            logger.critical(err)

        installed = re.findall(r"(\w+\.\w+):.*installed", proc.stdout)
        msg = f"Installed collections include: {oxford_join(installed)}"
        note(msg)

    def _install_local_collection(self: Installer) -> None:  # noqa: PLR0912, PLR0915
        """Install the collection from the build directory.

        Raises:
            RuntimeError: If tarball is not found or if more than one tarball is found.
        """
        msg = f"Installing local collection from: {self._config.collection_build_dir}"
        logger.info(msg)

        command = (
            "cp -r --parents $(git ls-files 2> /dev/null || ls)"
            f" {self._config.collection_build_dir}"
        )
        msg = "Copying collection to build directory using git ls-files."
        logger.debug(msg)
        try:
            subprocess_run(
                command=command,
                cwd=self._config.collection.path,
                verbose=self._config.args.verbose,
            )
        except subprocess.CalledProcessError as exc:
            err = f"Failed to copy collection to build directory: {exc} {exc.stderr}"
            logger.critical(err)

        command = (
            f"cd {self._config.collection_build_dir} &&"
            f" {self._config.venv_bindir / 'ansible-galaxy'} collection build"
            f" --output-path {self._config.collection_build_dir}"
            " --force"
        )

        msg = "Running ansible-galaxy to build collection."
        logger.debug(msg)

        try:
            subprocess_run(command=command, verbose=self._config.args.verbose)
        except subprocess.CalledProcessError as exc:
            err = f"Failed to build collection: {exc} {exc.stderr}"
            logger.critical(err)

        built = [
            f
            for f in Path(self._config.collection_build_dir).iterdir()
            if f.is_file() and f.name.endswith(".tar.gz")
        ]
        if len(built) != 1:
            err = (
                "Expected to find one collection tarball in"
                f"{self._config.collection_build_dir}, found {len(built)}"
            )
            raise RuntimeError(err)
        tarball = built[0]

        if self._config.site_pkg_collection_path.exists():
            msg = f"Removing installed {self._config.site_pkg_collection_path}"
            logger.debug(msg)
            if self._config.site_pkg_collection_path.is_symlink():
                self._config.site_pkg_collection_path.unlink()
            else:
                shutil.rmtree(self._config.site_pkg_collection_path)

        info_dirs = [
            entry
            for entry in self._config.site_pkg_collections_path.iterdir()
            if entry.is_dir()
            and entry.name.endswith(".info")
            and entry.name.startswith(self._config.collection.name)
        ]
        for info_dir in info_dirs:
            msg = f"Removing installed {info_dir}"
            logger.debug(msg)
            shutil.rmtree(info_dir)

        command = (
            f"{self._config.venv_bindir / 'ansible-galaxy'} collection"
            f" install {tarball} -p {self._config.site_pkg_path}"
            " --force"
        )
        env = {
            "ANSIBLE_GALAXY_COLLECTIONS_PATH_WARNING": str(self._config.args.verbose),
        }
        msg = "Running ansible-galaxy to install a local collection and it's dependencies."
        logger.debug(msg)
        try:
            proc = subprocess_run(
                command=command,
                env=env,
                verbose=self._config.args.verbose,
            )
        except subprocess.CalledProcessError as exc:
            err = f"Failed to install collection: {exc} {exc.stderr}"
            logger.critical(err)
            return

        # ansible-galaxy collection install does not include the galaxy.yml for version
        # nor does it create an info file that can be used to determine the version.
        # preserve the MANIFEST.json file for editable installs
        if not self._config.args.editable:
            shutil.copy(
                self._config.collection_build_dir / "galaxy.yml",
                self._config.site_pkg_collection_path / "galaxy.yml",
            )
        else:
            shutil.copy(
                self._config.site_pkg_collection_path / "MANIFEST.json",
                self._config.collection_cache_dir / "MANIFEST.json",
            )

        installed = re.findall(r"(\w+\.\w+):.*installed", proc.stdout)
        msg = f"Installed collections include: {oxford_join(installed)}"
        note(msg)

    def _swap_editable_collection(self: Installer) -> None:
        """Swap the installed collection with the current working directory.

        Raises:
            RuntimeError: If the collection path is not set.
        """
        msg = f"Swapping {self._config.collection.name} with {self._config.collection.path}"
        logger.info(msg)

        if self._config.collection.path is None:
            msg = "Collection path not set"
            raise RuntimeError(msg)
        msg = f"Removing installed {self._config.site_pkg_collection_path}"
        logger.debug(msg)
        if self._config.site_pkg_collection_path.exists():
            if self._config.site_pkg_collection_path.is_symlink():
                self._config.site_pkg_collection_path.unlink()
            else:
                shutil.rmtree(self._config.site_pkg_collection_path)

        msg = (
            f"Symlinking {self._config.site_pkg_collection_path}"
            f" to {self._config.collection.path}"
        )
        logger.debug(msg)
        self._config.site_pkg_collection_path.symlink_to(self._config.collection.path)

    def _pip_install(self: Installer) -> None:
        """Install the dependencies."""
        msg = "Installing python requirements."
        logger.info(msg)

        command = (
            f"{self._config.venv_interpreter} -m pip install"
            f" -r {self._config.discovered_python_reqs}"
        )

        msg = (
            f"Installing python requirements from {self._config.discovered_python_reqs}"
        )
        logger.debug(msg)
        try:
            subprocess_run(command=command, verbose=self._config.args.verbose)
        except subprocess.CalledProcessError as exc:
            err = (
                "Failed to install requirements from"
                f" {self._config.discovered_python_reqs}: {exc} {exc.stderr}"
            )
            logger.critical(err)
        else:
            msg = "All python requirements are installed."
            note(msg)

    def _check_bindep(self: Installer) -> None:
        """Check the bindep file."""
        msg = "Checking system packages."
        logger.info(msg)

        command = f"bindep -b -f {self._config.discovered_bindep_reqs}"
        try:
            subprocess_run(command=command, verbose=self._config.args.verbose)
        except subprocess.CalledProcessError as exc:
            lines = exc.stdout.splitlines()
            msg = (
                "Required system packages are missing."
                " Please use the system package manager to install them."
            )
            logger.error(msg)  # noqa: TRY400
            for line in lines:
                msg = f"Missing: {line}"
                logger.error(msg)  # noqa: TRY400
                pass
        else:
            msg = "All required system packages are installed."
            note(msg)
            return
